# calculation of sweet parameters based on the code of VG Scienta. This is querky and should be replaced in acquisition script by proper implementation (see sweep.py)
from .data_read import Metadata
import math

def get_sweep_parameters(spectrum_begin_energy: float, spectrum_end_energy: float, step_size: float, lens_k: float):
    """
    Extract sweep parameters (sweep start and end energy, channel width, steps number) based on implementation of VG Scienta.

    Args:
        spectrum_begin_energy: Begin energy of the spectrum.
        spectrum_end_energy: End energy of the spectrum.
        step_size: Width of the channel.
        lens_k: Lens configuration indicating the energy window width per measurement.

    Returns:
        4 sweep parameters (sweep start energy, sweep end energy, channel width, steps number).

    """
    # calculate adjusted channel width
    adjusted_channel_width = step_size / (1- lens_k/2)

    # calculate lower edge of first measurement (equals sweep start energy)
    spectrum_begin_upper_energy = spectrum_begin_energy / (1- lens_k/2) # caclulate upper edge of first measurement    
    lower_edge_e = (1 - lens_k) * spectrum_begin_upper_energy # calculate lower edge of first measurement
    channel_count = math.floor((spectrum_begin_upper_energy - lower_edge_e) / adjusted_channel_width) # calculate number of channels
    rest_e  = spectrum_begin_upper_energy - lower_edge_e - channel_count * adjusted_channel_width # calculate rest energy not convered by a full channel
    lower_edge_e += + rest_e - 1.5 * adjusted_channel_width # calculate new lower edge assuring each energy is hit by entire channel
    # continuesly increase lower edge until upper edge is below spectrum begin
    upper_edge_e = (lower_edge_e / (1 - lens_k/2)) * (1+ lens_k/2)
    while upper_edge_e <= spectrum_begin_upper_energy:
        lower_edge_e += adjusted_channel_width
        upper_edge_e = (lower_edge_e / (1 - lens_k/2)) * (1+ lens_k/2)
        noOfChannels = math.floor((upper_edge_e - lower_edge_e) / adjusted_channel_width)
        upper_edge_e = lower_edge_e + noOfChannels * adjusted_channel_width
    sweep_start_energy = lower_edge_e

    # calculate upper sweep end energy
    sweep_end_energy = spectrum_end_energy/(1 - lens_k/2)

    # calculate number of steps
    sweep_steps = math.floor((sweep_end_energy - sweep_start_energy) / adjusted_channel_width) + 1

    return sweep_start_energy, sweep_end_energy, adjusted_channel_width, sweep_steps, lens_k
    


import numpy as np

XYR_POINTS = 101

def get_default_xyr_config(spatial_diameter: int, scaling_factor: float = 1, radius: bool = False) -> list:
    """
    Extract the default x, y or r binning configuration.

    Args:
        spacial_diameter: Spacial diameter of the detector.
        scaling_factor: The scaling factor of the x, y r binning (default is 1). Important for x and y in SI units.
        radius: Wether the binning should be in radius or not.
    
    Returns:
        The default configuration for x, y or r binning configs.
    """
    xyr_min = 0 if radius else (-spatial_diameter/2)*scaling_factor
    xyr_max = (spatial_diameter/2)*scaling_factor
    return [xyr_min, xyr_max, XYR_POINTS]


T_UPSCALE_FACTOR = 10
T_MAX_POINTS = 201

def get_default_t_config(tof_vector: list, energy_matrix: list, e_ref: float, tdc_res: int, E_begin: float, E_end: float, t_0_tol: int, t0: int, si: bool = False) -> list:
    """
    Extract the default t binning configuration.

    Args:
        tof_vector: The time of flight vector from the acquisition.cfg.
        energy_matrix: The energy matrix from the acquisition.cfg.
        e_ref: The reference energy from the acquisition.cfg.
        tdc_res: The time to digital converter resolution from the acquisition.cfg.
        E_begin: The start energy from the acquisition.cfg.
        E_end: The end energy from the acquisition.cfg.
        t_0_tol: The tolerance for t_0 from the acquisition.cfg.
        t0: The t offset.
        si: Wether bin conf should be given in SI units (s).

    Returns:
        The default configuration for t binning configs.
    """
    ## calculate t0 for reference energy from energy matrix at theta=0
    # interpolate for higher precision
    t = np.linspace(tof_vector[0], tof_vector[-1], T_UPSCALE_FACTOR*len(tof_vector))
    E = np.interp(t, tof_vector, energy_matrix[0])
    # find crossing of reference energy and calculate t_0 in ticks
    crossing_idx = next(i for i,x in enumerate(E) if x < e_ref)
    t_0_ref = t[crossing_idx] * tdc_res
    # determin t_0 at spectrum start and end energy
    t_0_end = int(t_0_ref * np.sqrt(e_ref / E_begin ))
    t_0_begin = int(t_0_ref * np.sqrt(e_ref / E_end ))
    # calculate t_min und t_max based on t_0_tol
    t_min = (t_0_begin - t_0_tol + t0)
    t_max = (t_0_end + t_0_tol + t0)
    # calulate number of points
    t_points = min(T_MAX_POINTS, t_max - t_min)

    # if desired transform into seconds (si)
    if si:
        t_min /= tdc_res
        t_max /= tdc_res

    return [t_min, t_max, t_points]


PHI_MIN = -np.pi
PHI_MAX = np.pi
PHI_POINTS = 201

def get_default_phi_config() -> list:
    """
    Extract the default phi binning configuration.

    Returns:
        The default configuration for phi binning configs.
    """
    return [PHI_MIN, PHI_MAX, PHI_POINTS]

E_POINTS = 101

def get_default_E_config(E_min: float, E_max: float, E_step_size:float = None) -> list:
    """
    Extract the default E binning configuration.

    Args:
        E_min: The minimum value of the E binninge from aquisition.cfg.
        E_max: The maximum value of the E binninge from aquisition.cfg.
        E_step_size: The step size of the E binning e from aquisition.cfg for sweep mode. Default is None and there for E_POINTS will be used.

    Returns:
        The default configuration for E binning configs.
    """
    e_points = E_POINTS if E_step_size is None else int((E_max - E_min) / E_step_size) + 1

    return [E_min, E_max, e_points]

THETA_MIN = 0
THETA_POINTS = 201

def get_default_theta_config(theta_max: float) -> list:
    """
    Extract the default theta binning configuration.

    Args:
        theta_max: The maximum theta value from aquisition.cfg.

    Returns:
        The default configuration for theta binning configs.
    """
    return [THETA_MIN, theta_max, THETA_POINTS]


import numpy as np
from scipy.interpolate import RectBivariateSpline
from .data_read import Metadata
from .default_bin_config import get_default_xyr_config, get_default_t_config, get_default_phi_config, get_default_E_config, get_default_theta_config

class ARTOFTransformer:
    """Class to transform raw artof data"""

    def __init__(self, metadata: Metadata, x0: int = None, y0: int = None, t0: int = None):
        """Initializer ARTOFTransform class"""
        self.tdc_res = metadata.detector.tdcResolution
        self.e_kin_ref = metadata.lensmode.eKinRef
        self.lens_k = metadata.lensmode.lensK
        # create transformations
        self.x_transform, self.y_transform, self.t_transform = self.__ticks_to_SI_transform(metadata, x0, y0, t0)                
        self.E_transform, self.theta_transform = self.__tr_to_Etheta_transform(metadata)       
        # save t0 in tacks, t_min, and t_max in SI (at reference energy)
        self.t0 = metadata.detector.t0 if t0 is None else t0
        self.t_min, self.t_max = metadata.lensmode.tofVector[0], metadata.lensmode.tofVector[-1]
                    

    def transform(self, raw_data: list, load_as: str, center_energy: float = None) -> list:
        """
        Transform raw data to desired representation.

        Args:
            raw_data: 2D list of raw data points (x, y, t ticks).
            load_as: Desired representation to transform to (options: 'raw', 'raw_SI', 'cylindrical', 'spherical').

        Returns:
            Three 2D list of transformed data, list of variable names, and list of default bin edges for given transformation.
        """

        match load_as:
            case 'raw':
                data = raw_data
            case 'raw_SI':
                x, y, t = self.__raw_to_SI(raw_data)
                data = np.stack([x, y, t], -1)
            case 'cylindrical':
                r, phi, t = self.__raw_to_cylindrical(raw_data)
                data = np.stack([r, phi, t], -1)
            case 'spherical':           
                if center_energy is None:
                    raise ValueError('Center energy is required for spherical transformation.')             
                E, phi, theta = self.__raw_to_spherical(raw_data, center_energy)
                data = np.stack([E, phi, theta],-1)
            case _:
                print(f'Did not recognize transformation of type {load_as}. Using raw data')
                data = raw_data
        return data
    
    def __raw_to_spherical(self, raw_data: list, center_energy: float) -> list:
        """
        Transform raw data to spherical coordinates.

        Args:
            raw_data: 2D list of raw data points (x, y, t ticks).
            center_energy: Center energy in eV.
        
        Returns:
            3 lists containing E, phi, and theta values
        """
        energy_scaler = center_energy / self.e_kin_ref

        raw_data = self.__clip_t(raw_data, energy_scaler)
        r, phi, t = self.__raw_to_cylindrical(raw_data)
        t *= np.sqrt(energy_scaler)
        E = self.E_transform.ev(r, t) * energy_scaler
        theta = self.theta_transform.ev(r, t)
        E, phi, theta = self.__clip_E(E, phi, theta, center_energy)
        return E, phi, theta
    
    def __clip_t(self, raw_data, energy_scaler):    
        """
        Clip raw data to t_min and t_max in SI units.

        Args:
            raw_data: 2D list of raw data points (x, y, t ticks).
            energy_scaler: Energy scaler to scale energy and time.

        Returns:
            Clipped raw data.
        """    
        t_min_raw = self.t_min * self.tdc_res / np.sqrt(energy_scaler) + self.t0
        t_max_raw = self.t_max * self.tdc_res / np.sqrt(energy_scaler) + self.t0
        return raw_data[(raw_data[:,2] >= t_min_raw) & (raw_data[:,2] <= t_max_raw)]
    
    def __clip_E(self, E, phi, theta, center_energy):
        """
        Clip (E, phi, and theta) point values to be within lensK window.

        Args:
            E: Energy in eV.
            phi: Phi in radians.
            theta: Theta in radians.

        Returns:
            Clipped E, phi, and theta.
        """
        E_min = center_energy *  (1 - self.lens_k/2)
        E_max = center_energy *  (1 + self.lens_k/2)
        return E[(E >= E_min) & (E <= E_max)], phi[(E >= E_min) & (E <= E_max)], theta[(E >= E_min) & (E <= E_max)]
    
    def __raw_to_cylindrical(self, raw_data: list) -> list:
        """
        Transform raw data to cylindrical coordinates.

        Args:
            raw_data: 2D list of raw data points (x, y, t ticks).

        Returns:
            Three lists containing r, phi, and t values.
        """
        x, y, t = self.__raw_to_SI(raw_data)
        r, phi = self.__xy_to_polar(x,y)
        return r, phi, t

    def __raw_to_SI(self, raw_data: list) -> list:
        """
        Transform raw data to SI units.

        Args:
            raw_data: 2D list of raw data points (x, y, t ticks).

        Returns:
            Three lists containing x, y, and t values in SI units.
        """
        x = self.x_transform.ev(raw_data[:,1], raw_data[:,0])
        y = self.y_transform.ev(raw_data[:,1], raw_data[:,0])
        t = self.t_transform(raw_data[:,2])
        return x,y,t


    def get_axis_and_bins(self, load_as: str, metadata: Metadata, t0: int) -> tuple:
        """
        Get axis names and default bin edges for given transformation.

        Args:
            load_as: Desired representation to transform to (options: 'raw', 'raw_SI', 'cylindrical', 'spherical').
            metadata: Metadata class containing all metadata for current measurement.
            t_0: Time offset in ticks.
        """
        match load_as:
            case 'raw':
                x_bin_conf = get_default_xyr_config(metadata.detector.spatialDiameter)
                y_bin_conf = get_default_xyr_config(metadata.detector.spatialDiameter)
                t0 = metadata.detector.t0 if t0 is None else t0
                t_bin_conf = get_default_t_config(metadata.lensmode.tofVector, metadata.lensmode.energyMatrix, metadata.lensmode.eKinRef, 
                                                  metadata.detector.tdcResolution, metadata.general.spectrumBeginEnergy, metadata.general.spectrumEndEnergy,
                                                  metadata.detector.t0Tolerance, t0)

                return ['x_ticks', 'y_ticks', 't_ticks'], [x_bin_conf, y_bin_conf, t_bin_conf]
            case 'raw_SI':
                spatial_diameter = metadata.detector.spatialDiameter
                # implemented based on Igor ARTOFLoader, TODO check if correct
                x_vec_max = metadata.detector.transformXVector[-1]
                x_matrix_max = max([max(row) for row in metadata.detector.transformXMatrix])
                x_scaling = x_matrix_max / x_vec_max
                y_vec_max = metadata.detector.transformYVector[-1]            
                y_matrix_max = max([max(row) for row in metadata.detector.transformYMatrix])
                y_scaling = y_matrix_max / y_vec_max
                x_bin_conf = get_default_xyr_config(spatial_diameter, scaling_factor=x_scaling)
                y_bin_conf = get_default_xyr_config(spatial_diameter, scaling_factor=y_scaling)
                t0 = metadata.detector.t0 if t0 is None else t0
                t_bin_conf = get_default_t_config(metadata.lensmode.tofVector, metadata.lensmode.energyMatrix, metadata.lensmode.eKinRef, 
                                                  metadata.detector.tdcResolution, metadata.general.spectrumBeginEnergy, metadata.general.spectrumEndEnergy,
                                                  metadata.detector.t0Tolerance, t0, si=True)
                return ['x_m', 'y_m', 't_s'], [x_bin_conf, y_bin_conf, t_bin_conf]
            case 'cylindrical':
                spatial_diameter = metadata.detector.spatialDiameter
                # implemented based on Ugor ARTOFLoader, TODO check if correct
                x_vec_max = metadata.detector.transformXVector[-1]
                x_matrix_max = max([max(row) for row in metadata.detector.transformXMatrix])
                x_scaling = x_matrix_max / x_vec_max
                y_vec_max = metadata.detector.transformYVector[-1]            
                y_matrix_max = max([max(row) for row in metadata.detector.transformYMatrix])
                y_scaling = y_matrix_max / y_vec_max
                r_bin_conf = get_default_xyr_config(spatial_diameter, scaling_factor=max(x_scaling,y_scaling), radius=True)
                phi_bin_conf = get_default_phi_config()
                t0 = metadata.detector.t0 if t0 is None else t0
                t_bin_conf = get_default_t_config(metadata.lensmode.tofVector, metadata.lensmode.energyMatrix, metadata.lensmode.eKinRef,
                                                    metadata.detector.tdcResolution, metadata.general.spectrumBeginEnergy, metadata.general.spectrumEndEnergy,
                                                    metadata.detector.t0Tolerance, t0, si=True)
                return ['r_m', 'phi_rad', 't_s'], [r_bin_conf, phi_bin_conf, t_bin_conf]
            case 'spherical':
                E_min = metadata.general.spectrumBeginEnergy
                E_max = metadata.general.spectrumEndEnergy
                E_step_size = metadata.general.lensLowEdgeEnergyStep if metadata.general.acquisitionMode == 'sweep' else None
                E_bin_conf = get_default_E_config(E_min, E_max, E_step_size=E_step_size)
                phi_bin_conf = get_default_phi_config()
                theta_bin_conf = get_default_theta_config(metadata.lensmode.maxTheta)
                return ['E_eV', 'phi_rad', 'theta_rad'], [E_bin_conf, phi_bin_conf, theta_bin_conf] 
            case _:
                raise ValueError(f'Did not recognize transformation of type {load_as}.')

    def __ticks_to_SI_transform(self, metadata: Metadata, x0: int = None, y0: int = None, t0: int = None) -> tuple:
        """
        Transform x, y, and t from ticks to SI units using transformation matrices and tdcResolution from acquisition.cfg file.

        Args:
            metadata: Metadata class containing all metadata for current measurement.
            x0: x offset in ticks (default: from the acquisition.cfg file).
            t0: y offset in ticks (default: from the acquisition.cfg file).
            t0: t offset in ticks (default: from the acquisition.cfg file).

        Returns:
            Three lists containing x, y, and t values in SI units.
        """
        # convert x and y ticks to radius in m and phi in radians
        detector = metadata.detector
        x0 = detector.x0 if x0 is None else x0
        y0 = detector.y0 if y0 is None else y0
        x_transform = self.__create_matrix_transform(x0, y0, detector.transformXVector, detector.transformYVector, detector.transformXMatrix)
        y_transform = self.__create_matrix_transform(x0, y0, detector.transformXVector, detector.transformYVector, detector.transformYMatrix)

        # transform time ticks to time in seconds
        t_transform = lambda t:  self.__transform_time(t, self.t0, detector.tdcResolution)

        return x_transform, y_transform, t_transform


    def __xy_to_polar(self, x: float, y: float) -> tuple:    
        """
        Transform x and y in SI units to polar coordinates. The function arctan2(y, x) is used.

        Args:
            x: x value in meters (SI).
            y: y value in meters (SI).

        Returns:
            r in meters and phi in radians.
        """
        r = np.sqrt(x**2 + y**2)
        phi = np.arctan2(y, x)
        return r, phi


    def __tr_to_Etheta_transform(self, metadata: Metadata) -> tuple:
        """
        Transform t and r in SI units to E and theta. The transformation matrices from the acquisition.cfg file are used. Warning: Scaling has to be done when evaluting using factor centerEnergy/eKinRef.

        Args:
            metadata: Metadata class containing all metadata for current measurement.

        Returns:
            E in eV and theta in radians.
        """
        lensmode = metadata.lensmode
        t_vector = lensmode.tofVector
        r_vector = lensmode.radiusVector
        energy_matrix = lensmode.energyMatrix
        theta_matrix = lensmode.thetaMatrix

        E = self.__create_matrix_transform(0, 0, t_vector, r_vector, energy_matrix)
        theta = self.__create_matrix_transform(0, 0, t_vector, r_vector, theta_matrix)
        return E, theta


    def __transform_time(self, t_raw: int, t0: int, tdcResolution: float) -> float:
        """
        Transform time from ticks to seconds.

        Args:
            t_ticks: Time in ticks.
            t0: Time offset in ticks.
            tdcResolution: Resolutions of time to digital converter (tdc); number of events per second.

        Returns:
            Time in seconds.
        """
        return (t_raw - t0) * 1 / tdcResolution 


    def __create_matrix_transform(self, p1_0: int, p2_0: int, p1_vec: list, p2_vec: list, trans_mat: list) -> RectBivariateSpline:
        """
        Transform 2D data point using a given matrix using interpolation through a bivariate spline.

        Args:
            p1_0: Offset of p1.
            p2_0: Offset of p2.
            p1_vec: Vector corresponding to p1 and the columns of the matrix.
            p2_vec: Vector corresponding to p2 and the rows of the matrix.
            trans_mat: 2D list representing the transformation matrix.

        Returns:
            RectBivariateSpline interpolation for given matrix.
        """
        interp = RectBivariateSpline(p2_vec+p2_0, p1_vec+p1_0, trans_mat)
        return interp


import numpy as np
import matplotlib.pyplot as plt
import threading
from io import TextIOWrapper

# internal
from .data_read import read_metadata, load_file
from .data_process import get_bin_edges, project_data
from .Transformer import ARTOFTransformer
from .sweep_scienta import get_sweep_parameters
from .normalize import get_norm_scaling_factor, norm_sweep

# convert name of datatype to actual type
data_type_dict = {'int32': np.int32, 'float64': np.float64}


class ARTOFLoader:
    """Class to load ARTOF data."""

    def __init__(self):
        """Initialize ARTOFLoader class"""
        self.data = None
        self.binned_data = None
        self.bin_edges = list()
        self.format = None
        self.axes = None
        self.metadata = None
        self.transformer = None

    def load_run(self, path: str, format: str, x0: int = None, y0: int = None, t0: int = None, iter_interval: tuple = None, sweep_type: str = 'Sienta', multithreading=True):
        """
        Load ARTOF data for run in directory and transform into desired format.

        Args:
            path: Path to run directory.
            format: Load parameters in given format ('raw', 'raw_SI', 'cylindrical', 'spherical').
                `raw`: Load raw data in ticks (x,y,t).
                `raw_SI`: Load raw data in SI units (x,y,t).
                `cylindrical`: Load data in cylindrical coordinates (r, phi, t).
                `spherical`: Load data in spherical coordinates and associated energy (E, theta, phi).
            x0: Offset for x ticks, optional (default None).
            y0: Offset for y ticks, optional (default None).
            z0: Offset for t ticks, optional (default None).
            iter_interval: Tuple of start (including) and stop (excluding) lens iteration to load (default None, load all).
            sweep_type: Type of sweep analysis ('Sienta' or 'normal').
            multithreading: Use multithreading for data loading (default True).
        """
        # save format 
        self.format = format
        # aquire metadata
        self.metadata = read_metadata(path)
        # setup everything need for sweep analysis
        self.acquisitionMode = self.metadata.general.acquisitionMode
        if self.acquisitionMode == 'sweep':
            if sweep_type == 'Sienta':
                general = self.metadata.general
                self.sweep_start_energy, self.sweep_end_energy, self.adjusted_channel_width, self.lens_steps, self.lens_k = get_sweep_parameters(general.spectrumBeginEnergy,
                    general.spectrumEndEnergy, general.lensLowEdgeEnergyStep, self.metadata.lensmode.lensK)
            else:
                raise Exception('Normal sweep analysis not implemented yet.')
        else: 
            self.lens_steps = 1
        # create transformer based on metadata and transformation parameters
        self.transformer = ARTOFTransformer(self.metadata, x0, y0, t0)
        # save axes and set returned bin configurations as default
        self.axes, self.default_bin_confs = self.transformer.get_axis_and_bins(format, self.metadata, t0)

        # setup range of iterations to load
        if iter_interval is None:
            iter_range = range(self.metadata.general.lensIterations)
        else:
            if iter_interval[1] > self.metadata.general.lensIterations or iter_interval[0] < 0:
                raise ValueError(f'Given range of iterations is not within range of available iterations (0 to {self.metadata.general.lensIterations-1}).')
            iter_range = range(iter_interval[0], iter_interval[1])
        # save number of iterations
        self.iterations = len(iter_range)
        # setup progress information                
        self.progress_info = {'current': 0, 'total': len(iter_range) * self.lens_steps}
        self.__print_progress()
        # transform data to desired format via multithreading if desired
        data_pieces = []
        if multithreading:
            threads = []
            for iter in iter_range:
                for step in range(self.lens_steps):                    
                    thread = threading.Thread(target=self.__process_data, args=(path, iter, step, data_pieces, format))
                    threads.append(thread)
                    thread.start()
            for thread in threads:
                thread.join()
        else:
            for iter in range(self.metadata.general.lensIterations):
                for step in range(self.lens_steps):
                    self.__process_data(path, iter, step, data_pieces, format)
        self.data = np.concatenate(data_pieces, axis=0)

        # print information about loaded data
        print() # need to stop overwriting of progress bar
        print(f'Loaded and transformed {self.data.shape[0]} data points to formats {self.axes}.')

    def save_transformed_data(self, path: str):
        """
        Save transformed data to binary file (path + '.bin') and save addional info file (path + '_info.txt')

        Args:
            path: Path to file where transformed data should be stored.
        """        
        self.data.tofile(f'{path}.bin') 
        with open(f'{path}_info.txt', 'w') as f:
            self.__write_par(f, type(self.data[0,0]).__name__)
            self.__write_par(f, self.format)
            self.__write_par(f, self.acquisitionMode)
            if self.acquisitionMode == 'sweep':
                self.__write_par(f, f'{self.sweep_start_energy},{self.sweep_end_energy},{self.adjusted_channel_width},{self.lens_steps},{self.lens_k}')
            self.__write_par(f, ",".join(self.axes))
            self.__write_par(f, ";".join(map(str, self.default_bin_confs)))

        print(f'Saved transformed data as binary file to {path}_{self.format}.bin and additional information to {path}_{self.format}_info.txt')


    def __write_par(self, file: TextIOWrapper, par):
        """
        Write parameter to file with new line at the end.

        Args:
            file: File to write to.
            par: Parameter to write.
        
        """
        file.write(par)
        file.write('\n')

    def load_transformed_data(self, data_path: str, info_path: str):
        """
        Load transformed data from file.

        Args:
            data_path: Path to file where transformed data is stored.
            info_path: Path to file where information about the transformed data is stored.
        """
        with open(info_path) as f:
            data_type = data_type_dict[next(f).strip()]
            self.format = next(f).strip()
            self.acquisitionMode = next(f).strip()
            if self.acquisitionMode == 'sweep':
                self.sweep_start_energy, self.sweep_end_energy, self.adjusted_channel_width, self.lens_steps, self.lens_k = map(float, next(f).strip().split(','))
                self.lens_steps = int(self.lens_steps)
            self.axes = next(f).strip().split(',')
            self.default_bin_confs = [eval(bin_conf) for bin_conf in next(f).strip().split(';')]
        self.data = np.fromfile(data_path, dtype=data_type).reshape(-1,3)


    def bin_data(self, bin_confs: list = [None, None, None], norm_modes: list = None):
        """
        Bin loaded data into 3D histogram.

        Args:        
            bin_confs: List of 3 binning configurations for the 3 parameters [min, max, points]. F.e.: [[-1500, 1500, 101], [-1500, 1500, 101],[12000, 18000, 201]].       
            norm_mode: Normalization mode for binned data ('iterations', 'dwell_time', 'step_size',  'sweep'). Default is None.
                `iterations`: Normalize data by number of iterations.
                `dwell_time`: Normalize data by dwell time.
                `sweep`: Normalize data by changing window size of sweep data.

        Raises:
            Exception: If data is not loaded before binning.
        """
        if self.data is None:
            raise Exception('Load the data before binning the data.')

        # use either passed or default bin configurations
        cur_bin_confs = []
        for i, bin_conf in enumerate(bin_confs):
            if bin_conf is None:
                print(f'Using default bin configuration for {self.axes[i]}: {self.default_bin_confs[i]}')
                cur_bin_confs.append(self.default_bin_confs[i])
            else:
                cur_bin_confs.append(bin_conf)
        
        # create bin edges based on the passed bin configs
        self.bin_edges = list()
        for i in range(3):
            self.bin_edges.append(get_bin_edges(self.data[:,i], cur_bin_confs[i], data_id=self.axes[i]))
        # bin data in 3D histogram
        self.binned_data, _ = np.histogramdd(self.data, bins=self.bin_edges) 

        # normalize data if desired        
        if norm_modes is not None:
            scaling_factor = get_norm_scaling_factor(norm_modes, self.iterations, self.metadata.general.lensDwellTime)
            self.binned_data *= scaling_factor
            # normalize data by sweep acceptance if desired and in sweep mode
            if 'sweep' in norm_modes: 
                if self.acquisitionMode != 'sweep' or self.format != 'spherical':
                    raise Exception('Sweep normalization only possible for sweep data in spherical format.')
                self.binned_data = norm_sweep(self.binned_data, self.bin_edges[0], self.sweep_start_energy,
                                              self.adjusted_channel_width, self.lens_steps, self.lens_k)
                
            # print all non-recognized norm modes
            for mode in norm_modes:
                if mode not in ['iterations', 'dwell_time', 'sweep']:
                    print(f'Normalization mode "{mode}" not recognized.')
            
            
                
    

    def __process_data(self, path: str, iter: int, step: int, data_pieces: list, load_as: str):        
        """
        Load and transform single data file in given format (needed for multithreading).

        Args:
            path: Path where data files are located.
            iter: Index of the lens iteration to be loaded.
            step: Index of the lens step to be loaded.
            data_pieces: List of transformed data pieces to which the newly transformed data should be appended.
            load_as: Desired transformation format.
        """
        raw_data = load_file(path, iter, step)
        if self.acquisitionMode == 'sweep':
            center_energy = self.sweep_start_energy + step * self.adjusted_channel_width
            data_pieces.append(self.transformer.transform(raw_data, load_as, center_energy=center_energy))
        else:
            center_energy = self.metadata.general.centerEnergy
            data_pieces.append(self.transformer.transform(raw_data, load_as, center_energy=center_energy))
        self.progress_info['current'] += 1
        self.__print_progress()


    def __print_progress(self):
        """
        Print progress information.
        """
        current = self.progress_info['current']
        total = self.progress_info['total']
        print('\r', end='')
        print(f'Progress: [{"="*int(current*20/total):<20}] {current}/{total}', end='\r')


    def plot(self, axes: list, ranges: list = [None, None, None], norm_step_size: bool = False, width: float = 5.5, height: float = 5.5):
        """
        Plot loaded data as projection onto given axes. Projections are possible onto 1 or 2 axes.

        Args:
            axes: List containing all axes onto which the projection is performed, e.g., [0,1].
            ranges: List containing ranges for axes (e.g., [[50, 101], [0,50], None]), if None entire range of axes is used (default entire range of each axis).
            norm_step_size: Normalize data with step size before plotting (default False).
            width: Width of plot (default 5).
            height: Height of plot (default 5).
        """
        
        proj_data = project_data(self.binned_data, self.bin_edges, axes, ranges, norm_step_size)

        if len(axes) == 2: # plot data in 2D as image
            img_fig, img_ax = plt.subplots(figsize=(width, height))
            img_fig.subplots_adjust(bottom=0.2, left=0.2)
            img_ax.imshow(proj_data, cmap='terrain', norm='linear', origin='lower', extent=[self.bin_edges[axes[0]][0], self.bin_edges[axes[0]][-1], self.bin_edges[axes[1]][0], self.bin_edges[axes[1]][-1]], aspect="auto")
            img_ax.set_xlabel(self.__axis_label(self.axes[axes[0]]))
            img_ax.set_ylabel(self.__axis_label(self.axes[axes[1]]))
        elif len(axes) == 1: # plot data in 1D as line
            x_values = [(self.bin_edges[axes[0]][i] + self.bin_edges[axes[0]][i+1])/2 for i in range(len(self.bin_edges[axes[0]])-1)]            

            line_fig, line_ax = plt.subplots(figsize=(width, height))
            line_fig.subplots_adjust(bottom=0.2, left=0.2)
            line_ax.plot(x_values, proj_data)
            line_ax.set_xlabel(self.__axis_label(self.axes[axes[0]]))
            line_ax.set_ylabel('Counts')


        else:
            raise Exception(f'A projection along {len(axes)} axes is not possible.')
        
    def export_to_csv(self, path: str, axes: list, ranges: list = [None, None, None], norm_step_size: bool = False, delimiter: str = ','):
        """
        Export loaded data as projection onto given axes. Projections are possible onto 1 or 2 axes.

        Args:
            path: Path including file name to which the data is saved.
            axes: List containing all axes onto which the projection is performed, e.g., [0,1].
            ranges: List containing ranges for axes (e.g., [[50, 101], [0,50], None]), if None entire range of axes is used (default entire range of each axis).
            norm_step_size: Normalize data with step size before plotting (default False).
            delimiter: Delimiter by which the data is separated (default ',').
        """  
        data_to_export = project_data(self.binned_data, self.bin_edges, axes, ranges, norm_step_size)        
        np.savetxt(path, data_to_export, delimiter=delimiter)

    def __axis_label(self, axis: str) -> str:
        """
        Build string for matplotlib axis label including Greek characters.

        Args:
            axis: String containing the axis label and unit separated by '_'.

        Returns:
            Formatted string for matplotlib.
        """
        name, unit = axis.split('_')
        match name:
            case 'phi':
                name = '$\\varphi$'
            case 'theta':
                name = '$\\theta$'
        return f'{name} [{unit}]'






import numpy as np

def get_norm_scaling_factor(norm_modes: list, iterations: int, dwell_time: float) -> float:
    """
    Get scaling factor for normalization based on normalization modes.

    Args:
        norm_modes: List of normalization modes ('iterations', 'dwell time').            
        iterations: Number of iterations.
        dwell_time: Dwell time.

    Returns:
        Scaling factor for normalization.
    """
    scaling_factor = 1
    if 'iterations' in norm_modes:
        scaling_factor /= iterations
    if 'dwell_time' in norm_modes: 
        scaling_factor /= (dwell_time/1000) # norm to seconds
    return scaling_factor

def norm_sweep(binned_data: np.ndarray, bin_edges: list, sweep_start: float, step_size: float, sweep_steps: int, lens_k: float) -> np.ndarray:
    """
    Normalize sweep data to lens k value.

    Args:
        binned_data: Binned data to normalize.
        energy_bin_edges: Energy bin edges.
        sweep_start: Start value of sweep.
        step_size: Step size of sweep.
        sweep_steps: Number of steps in sweep.
        lens_k: Lens k value.

    Returns:
        Normalized sweep data.
    """
    # calculate bin centers
    bin_centers = np.array([(bin_edges[i] + bin_edges[i+1]) / 2 for i in range(len(bin_edges)-1)])
    center_hits = np.zeros(len(bin_centers))
    # count how often each bin center is hit during
    for i in range(sweep_steps):
        e = sweep_start + i * step_size
        step_min = e*(1-lens_k/2)
        step_max = e*(1+lens_k/2)
        hits = np.where(np.logical_and(bin_centers >= step_min, bin_centers <= step_max))    
        for j in hits[0]:
            center_hits[j] += 1
    # fit to linear function
    fit_pars = np.polyfit(bin_centers, center_hits, 1)
    norm_values = np.polyval(fit_pars, bin_centers)
    # normalize data
    return binned_data / norm_values[:, np.newaxis, np.newaxis]


import numpy as np
    
def get_bin_edges(data: list, bin_config: list, data_id: str = "unknown") -> np.ndarray:
    """
    Create bin edges (limits) for given bin config.

    Args:
        data: 1D list of data values.
        bin_config: Bin config consisting of three values: [min, max, points].
        data_id: Identifier to raise error message with identifier.

    Returns:
        List of bin edges.
    """
    min_point, max_point, total_points = bin_config
    if type(total_points) != int:
        raise Exception(f"The number of bins for {data_id} needs given as an int")

    bin_size = (max_point - min_point) / (total_points-1)
    # create bin edges in a way that bin centers are min and max
    return np.linspace(min_point - bin_size/2, max_point + bin_size/2, total_points+1)



    
def project_data(data: np.ndarray, bin_edges: list, axes: list, ranges: list, norm_step_size: bool) -> np.ndarray:
    """
    Plot loaded data in given projection. Projections are possible in 1 or 2 dimensions.

    Args:
        data: 3D data to be transformed.
        bin_edges: List of bin edges for each axis.
        axes: List containing all axes onto which the projection is performed, e.g., [0,1].
        ranges: List of ranges for each axis.
        norm_step_size: Normalize projection with step size.

    Returns:
        1 or 2D list containing the projection.
    """

    if data is None:
        raise Exception('Load and bin the data before projecting the data.')         
    if len(axes) not in [1,2]:
        raise Exception(f'A projection onto {len(axes)} axes is not possible.')
    
    # determine ranges (None is entire range)
    start_1, end_1 = (0, data.shape[0]) if ranges[0] is None else (ranges[0][0], ranges[0][1])
    start_2, end_2 = (0, data.shape[1]) if ranges[1] is None else (ranges[1][0], ranges[1][1])
    start_3, end_3 = (0, data.shape[2]) if ranges[2] is None else (ranges[2][0], ranges[2][1])

    proj_data = data[start_1:end_1,start_2:end_2,start_3:end_3]
    
    if len(axes) == 2: # project data onto 2 axes
        # determine the projection axis
        proj_axis = 3 - sum(axes)
        # switch order of array if needed
        if axes[0] < axes[1]:
            proj_data = np.swapaxes(proj_data, axes[0], axes[1])
        proj_data = proj_data[:,:,:].sum(axis=proj_axis)
    elif len(axes) == 1: # project data onto 1 axis
        # get axes to project along
        proj_axes = [0,1,2]
        proj_axes.remove(axes[0])
        # project along axes
        proj_data = proj_data.sum(axis=tuple(proj_axes))

    # normalize data with step size if desired
    if norm_step_size:
        step_size_product = 1
        for i in axes:
            step_size_product *= bin_edges[i][1] - bin_edges[i][0]
        proj_data /= step_size_product

    return proj_data